import logging
import typing as tp
from .base import EmbeddedSubmetrics, LeafMetric
from .registry import register_metric
logger = logging.getLogger(__name__)


@register_metric
class CounterMetric(EmbeddedSubmetrics):
    """
    A counter that can be adjusted by a given value.

    :param sum_children: whether to sum up all calls to children
    :param count_calls: count the amount of calls to handle()
    """
    CLASS_NAME = 'counter'

    def __init__(self, name, root_metric: 'Metric' = None, metric_level: str = None,
                 sum_children: bool = True,
                 count_calls: bool = False, *args, **kwargs):
        super().__init__(name, root_metric, metric_level, sum_children=sum_children,
                         count_calls=count_calls, *args, **kwargs)
        self.sum_children = sum_children
        self.count_calls = count_calls
        self.calls = 0
        self.value = 0

    def to_json(self) -> list:
        if self.embedded_submetrics_enabled:
            k = super().to_json()
            if self.sum_children:
                r = LeafMetric.to_json(self)
                r['_'] = self.value
                k = {'_': k, 'sum': r}
                if self.count_calls:
                    q = LeafMetric.to_json(self)
                    q['_'] = self.calls
                    k['count'] = q
            return k

        p = super().to_json()
        p['_'] = self.value
        if self.count_calls:
            r, q = super().to_json(), super().to_json()
            r['_'] = self.value
            p['_'] = r
            q['_'] = self.calls
            p['count'] = q
        return p

    def _handle(self, delta: float = 0, **labels):
        if self.embedded_submetrics_enabled or labels:
            if self.sum_children:
                self.value += delta
            self.calls += 1
            return super()._handle(delta, **labels)

        self.value += delta
        self.calls += 1



import logging
import io
import copy
from satella.coding import for_argument
from satella.instrumentation.metrics.json import get_labels_for_node

from ..json import is_leaf_node

logger = logging.getLogger(__name__)


class RendererObject(io.StringIO):

    @for_argument(None, copy.copy, None, copy.copy)
    def render_node(self, tree, prefixes, labels):
        if isinstance(tree, list):
            for elem in tree:
                self.render_node(elem, prefixes, labels)
            return

        if is_leaf_node(tree):
            self.write('_'.join(prefix for prefix in prefixes if prefix != ''))
            main_value = tree.pop('_')
            tree.update(labels)
            ts = tree.pop('_timestamp', None)
            curly_braces_used = len(tree) > 0
            if curly_braces_used:
                self.write('{')
            if curly_braces_used:
                labels = []
                for key, value in tree.items():
                    value = str(value).replace('\\', '\\\\').replace('"', '\\"').replace("'", '"')
                    labels.append('%s="%s"' % (key, value))
                self.write(','.join(labels))
                self.write('}')
            self.write(' %s' % (repr(main_value), ))
            if ts is not None:
                self.write(' %s' % (int(ts*1000), ))
            self.write('\n')

        else:
            labels = get_labels_for_node(tree)
            for k in labels:
                del tree[k]

            for k, v in tree.items():
                if k == '_' and isinstance(v, list):
                    for elem in v:
                        self.render_node(elem, prefixes, labels)
                    continue
                elif k == '_timestamp':
                    continue
                elif k != '_':
                    new_prefixes = prefixes + [k]
                else:
                    new_prefixes = prefixes
                self.render_node(v, new_prefixes, labels)


def json_to_prometheus(tree) -> str:
    """
    Render the JSON in the form understandable by Prometheus

    :param tree: JSON returned by the root metric (or any metric for that instance).
    :return: a string output to present to Prometheus
    """
    if not tree:
        return '\n'
    obj = RendererObject()
    obj.render_node(tree, [], {})
    return obj.getvalue().replace('\r\n', '\n')



import collections
import functools
import logging
import time
import typing as tp

import math

from .base import EmbeddedSubmetrics, RUNTIME, LeafMetric
from .registry import register_metric

logger = logging.getLogger(__name__)


# shamelessly taken from http://code.activestate.com/recipes/511478-finding-the-percentile-of-the-values/)
def percentile(n: tp.List[float], percent: float) -> float:
    """
    Find the percentile of a list of values.

    :param n: - is a list of values. Note this MUST BE already sorted.
    :param percent: - a float value from 0.0 to 1.0.

    :return: the percentile of the values
    """
    k = (len(n) - 1) * percent
    f = math.floor(k)
    c = math.ceil(k)
    if f == c:
        return n[int(k)]
    d0 = n[int(f)] * (c - k)
    d1 = n[int(c)] * (k - f)
    return d0 + d1


@register_metric
class QuantileMetric(EmbeddedSubmetrics):
    """
    A metric that can register some values, sequentially, and then calculate quantiles from it

    :param last_calls: last calls to handle() to take into account
    :param quantiles: a sequence of quantiles to return in to_json
    :param aggregate_children: whether to sum up children values (if present)
    :param count_calls: whether to count total amount of calls and total time
    """

    CLASS_NAME = 'quantile'

    def __init__(self, name, root_metric: 'Metric' = None, metric_level: str = None,
                 last_calls: int = 100, quantiles: tp.Sequence[float] = (0.5, 0.95),
                 aggregate_children: bool = True,
                 count_calls: bool = True, *args,
                 **kwargs):
        super().__init__(name, root_metric, metric_level, *args, last_calls=last_calls,
                         quantiles=quantiles, aggregate_children=aggregate_children,
                         count_calls=count_calls,
                         **kwargs)
        self.last_calls = last_calls
        self.calls_queue = collections.deque()
        self.quantiles = quantiles
        self.aggregate_children = aggregate_children
        self.count_calls = count_calls
        self.tot_calls = 0
        self.tot_time = 0

    def _handle(self, time_taken: float, **labels) -> None:
        if self.count_calls:
            self.tot_calls += 1
            self.tot_time += time_taken

        if labels or self.embedded_submetrics_enabled:
            return super()._handle(time_taken, **labels)
        if len(self.calls_queue) == self.last_calls:
            self.calls_queue.pop()
        self.calls_queue.appendleft(time_taken)

    def to_json(self):
        k = self._to_json()
        if self.count_calls:
            if isinstance(k, list):
                return {'count': {'_': self.tot_calls}, 'total': {'_': self.tot_time}, '_': k}
            else:
                k['count'] = {'_': self.tot_calls}
                k['total'] = {'_': self.tot_time}
                if self.enable_timestamp:
                    k['total']['_timestamp'] = k['count']['_timestamp'] = self.last_updated

                return k
        return k

    def _to_json(self) -> dict:
        if self.embedded_submetrics_enabled:
            k = super().to_json()
            if not self.aggregate_children:
                return k
            total_calls = []
            for child in self.children:
                total_calls.extend(child.calls_queue)
            total_calls.sort()
            k = {
                '_': k,
                'sum': self.calculate_quantiles(total_calls)
            }
            if self.enable_timestamp:
                k['_timestamp'] = self.last_updated
            return k
        else:
            return self.calculate_quantiles(self.calls_queue)

    def calculate_quantiles(self, calls_queue):
        output = []
        sorted_calls = sorted(calls_queue)
        for p_val in self.quantiles:
            k = LeafMetric.to_json(self)
            if not sorted_calls:
                k.update(quantile=p_val, _=0.0)
            else:
                k.update(quantile=p_val,
                         _=percentile(sorted_calls, p_val))
            if self.enable_timestamp:
                k['_timestamp'] = self.last_updated
            output.append(k)
        return output

    def measure(self, include_exceptions: bool = True, logging_level: int = RUNTIME,
                value_getter: tp.Callable[[], float] = time.monotonic, **labels):
        """
        A decorator to measure a difference between some value after the method call
        and before it.

        By default, it will measure the execution time.

        Use like:

        >>> call_time = getMetric('root.metric_name.execution_time', 'percentile')
        >>> @call_time.measure()
        >>> def measure_my_execution(args):
        >>>     ...

        :param include_exceptions: whether to include exceptions
        :param logging_level: one of RUNTIME or DEBUG
        :param value_getter: a callable that takes no arguments and returns a float, which is
            the value
        :param labels: extra labels to call handle() with
        """

        def outer(fun):
            @functools.wraps(fun)
            def inner(*args, **kwargs):
                start_value = value_getter()
                excepted = None
                try:
                    return fun(*args, **kwargs)
                except Exception as e:
                    excepted = e
                finally:
                    value_taken = value_getter() - start_value
                    if excepted is not None and not include_exceptions:
                        raise excepted

                    self.handle(logging_level, value_taken, **labels)

                    if excepted is not None:
                        raise excepted

            return inner

        return outer
